from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic
from wq.db.patterns.base import SerializableGenericRelation
from wq.db.patterns.base.models import NaturalKeyModelManager, NaturalKeyModel
from django.template.defaultfilters import slugify


class IdentifierManager(models.Manager):
    _cache = {}

    @classmethod
    def update_cache(cls, ident):
        key = '%s-%s' % (ident.content_type_id, ident.object_id)
        cls._cache[key] = ident

    def get_for_object(self, obj):
        ct = ContentType.objects.get_for_model(obj)
        key = '%s-%s' % (ct.pk, obj.pk)
        cache = type(self)._cache
        if key in cache:
            return cache[key]
        ids = self.filter(content_type=ct, object_id=obj.pk, is_primary=True)
        if ids.count() > 0:
            cache[key] = ids[0]
            return cache[key]
        return None

    def filter_by_identifier(self, identifier):
        return self.filter(
            models.Q(slug=identifier) | models.Q(name=identifier)
        ).order_by('-is_primary')

    def resolve(self, identifiers, exclude_apps=[]):
        resolved = None
        unresolved = None
        for identifier in identifiers:
            ids = self.filter_by_identifier(identifier)
            if exclude_apps:
                ids = ids.exclude(content_type__app_label__in=exclude_apps)
            items = ids.distinct('content_type', 'object_id').count()
            if items == 1:
                if resolved is None:
                    resolved = {}
                resolved[identifier] = ids[0]
            else:
                if unresolved is None:
                    unresolved = {}
                unresolved[identifier] = ids
        return resolved, unresolved

    # Default implementation of get_or_create doesn't work well with generics
    def get_or_create(self, **kwargs):
        try:
            return self.get(**kwargs), False
        except self.model.DoesNotExist:
            return self.create(**kwargs), True

    def find_unique_slug(self, name, model):
        if len(name) > 45:
            name = name[:45]
        slug = slugify(name)
        exists = self.filter(
            content_type__name=model,
            slug=slug
        )
        num = ''
        while exists.count() > 0:
            slug = slugify('%s %s' % (name, num))
            exists = self.filter(
                content_type__name=model,
                slug=slug
            )
            if num == '':
                num = 1
            else:
                num += 1
        return slug


class Identifier(models.Model):
    name = models.CharField(max_length=255, db_index=True)
    slug = models.SlugField()
    authority = models.ForeignKey('Authority', blank=True, null=True)
    is_primary = models.BooleanField(default=False)

    # Identifer can contain a pointer to any model
    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    content_object = generic.GenericForeignKey()

    objects = IdentifierManager()

    @property
    def url(self):
        if (not self.authority or not self.authority.object_url):
            return None
        else:
            return self.authority.object_url % self.slug

    def save(self, *args, **kwargs):
        if self.slug is None or self.slug == '':
            model = self.content_type.name
            self.slug = type(self).objects.find_unique_slug(self.name, model)
        super(Identifier, self).save(*args, **kwargs)
        if self.is_primary:
            IdentifierManager.update_cache(self)

    def __unicode__(self):
        return self.name

    class Meta:
        db_table = 'wq_identifier'


class PrimaryIdentifierManager(IdentifierManager):
    use_for_related_fields = True

    def get_query_set(self):
        qs = super(PrimaryIdentifierManager, self).get_query_set()
        return qs.filter(is_primary=True)


class PrimaryIdentifier(Identifier):
    objects = PrimaryIdentifierManager()

    class Meta:
        proxy = True


class IdentifiedModelManager(NaturalKeyModelManager):
    def get_by_identifier(self, identifier, auto_create=False):
        searches = [
            {'identifiers__slug': identifier, 'identifiers__is_primary': True},
            {'identifiers__name': identifier, 'identifiers__is_primary': True},
            {'identifiers__slug': identifier},
            {'identifiers__name': identifier},
            {'pk': identifier}
        ]

        object = None
        for search in searches:
            try:
                object = self.get(**search)
                break
            except (ValueError, self.model.DoesNotExist):
                if 'pk' not in search:
                    continue
                elif (not auto_create):
                    name = self.model._meta.object_name
                    raise self.model.DoesNotExist(
                        '%s "%s" does not exist' % (name, identifier)
                    )

        if object is None and auto_create:
            object = self.create_by_natural_key(identifier)

        return object

    def get_by_natural_key(self, identifier):
        return self.get_by_identifier(identifier)

    def create_by_natural_key(self, identifier):
        if ('name' in self.model._meta.get_all_field_names()):
            object = self.create(name=identifier)
        else:
            object = self.create()
        object.identifiers.create(name=identifier, is_primary=True)
        return object

    def get_query_set(self):
        qs = super(IdentifiedModelManager, self).get_query_set()
        meta = self.model._meta
        if meta.ordering:
            return qs
        ct = ContentType.objects.get_for_model(self.model)
        query = (
            """ SELECT name FROM wq_identifier
                WHERE content_type_id=%s AND object_id=%s.%s AND is_primary
                LIMIT 1 """
            % (ct.pk, meta.db_table, meta.pk.get_attname_column()[1]))
        qs = qs.extra(select={'wq_id_name': query})
        return qs.order_by('wq_id_name')


class IdentifiedModel(NaturalKeyModel):
    identifiers = SerializableGenericRelation(Identifier)
    primary_identifiers = generic.GenericRelation(
        PrimaryIdentifier, related_name='%(app_label)s_%(class)s_primary'
    )
    objects = IdentifiedModelManager()

    @classmethod
    def get_natural_key_fields(cls):
        return ['primary_identifiers__slug']

    @property
    def primary_identifier(self):
        return Identifier.objects.get_for_object(self)

    def fallback_identifier(self):
        if hasattr(self, 'name'):
            return unicode(self.name)
        else:
            return ContentType.objects.get_for_model(self).name

    def __unicode__(self):
        if self.primary_identifier:
            return self.primary_identifier.name
        else:
            return self.fallback_identifier()

    def natural_key(self):
        if self.primary_identifier:
            return (self.primary_identifier.slug,)
        else:
            return (self.pk,)

    class Meta:
        abstract = True


class Authority(models.Model):
    name = models.CharField(max_length=255)
    homepage = models.URLField(null=True, blank=True)
    object_url = models.URLField(null=True, blank=True)

    class Meta:
        verbose_name_plural = 'authorities'
        db_table = 'wq_identifiertype'

    def __unicode__(self):
        return self.name



from rest_pandas import PandasView
from wq.db.patterns.models import Identifier
from .serializers import EventResultSerializer
import swapper
EventResult = swapper.load_model('vera', 'EventResult')


class ChartView(PandasView):
    model = EventResult
    serializer_class = EventResultSerializer
    ignore_extra = True

    def get_queryset(self):
        qs = super(ChartView, self).get_queryset()
        qs = qs.select_related('event_site', 'result_type')
        return qs

    @property
    def filter_options(self):
        if hasattr(self, '_filter_options'):
            return self._filter_options

        slugs = self.kwargs['ids'].split('/')
        id_map, unresolved = Identifier.objects.resolve(
            slugs, exclude_apps=['dbio']
        )
        options = {}
        if unresolved:
            options['extra'] = []
            for key, items in unresolved.items():
                if len(items) > 0:
                    raise Exception(
                        "Could not resolve %s to a single item!" % key
                    )
                options['extra'].append(key)

        if id_map:
            for slug, ident in id_map.items():
                ctype = ident.content_type.model
                if ctype not in options:
                    options[ctype] = []
                options[ctype].append(ident)

        self._filter_options = options
        return options

    def filter_queryset(self, qs):
        for name, idents in self.filter_options.items():
            if name == "extra":
                continue
            fn = getattr(self, 'filter_by_%s' % name, None)
            ids = [ident.object_id for ident in idents]
            if fn:
                qs = fn(qs, ids)
            else:
                raise Exception("Don't know how to filter by %s" % name)
        if 'extra' in self.filter_options:
            qs = self.filter_by_extra(qs, self.filter_options['extra'])
        return qs

    def filter_by_site(self, qs, ids):
        return qs.filter(event_site__in=ids)

    def filter_by_parameter(self, qs, ids):
        return qs.filter(result_type__in=ids)

    def filter_by_extra(self, qs, extra):
        if self.ignore_extra:
            return qs
        else:
            raise Exception("Extra URL options found: %s" % extra)


class TimeSeriesMixin(object):
    """
    For use with chart.timeSeries() in wq/chart.js
    """

    def transform_dataframe(self, df):
        """
        The dataframe is already in a timeseries format.
        """
        return df


class ScatterMixin(object):
    """
    For use with chart.scatter() in wq/chart.js
    """

    def transform_dataframe(self, df):
        """
        Transform timeseries dataframe into a format suitable for plotting two
        values against each other.
        """

        # Only use primary 'value' column, ignoring any other result fields
        # that may have been added to a serializer subclass.
        for key in df.columns.levels[0]:
            if key != 'value':
                df = df.drop(key, axis=1)

        # Remove all indexes/columns except for parameter (which will become
        # the new 'value' field) and site (to allow distinguishing between
        # scatterplot data for each site).
        for name in df.columns.names:
            if name not in ("parameter", "site"):
                df.columns = df.columns.droplevel(name)

        # Rename columns ('value'/parameter column should be nameless)
        df.columns.names = ["", "site"]

        # Only include dates that have data for all parameters
        df = df.dropna(axis=0, how='any')
        return df


class BoxPlotMixin(object):
    """
    For use with chart.boxplot() in wq/chart.js
    """

    NAME_MAP = {
        'q1': 'p25',
        'q3': 'p75',
        'med': 'median',
        'whishi': 'max',
        'whislo': 'min',
    }

    def transform_dataframe(self, df):
        """
        Use matplotlib to compute boxplot statistics on timeseries data.
        """
        from pandas import DataFrame
        group = self.get_grouping(len(df.columns))
        if "index" in group:
            # Separate stats for each column in dataset
            groups = {
                col: df[col]
                for col in df.columns
            }
        else:
            # Stats for entire dataset
            df = df.stack().stack().stack()
            df.reset_index(inplace=True)
            index = self.get_serializer().get_index(df)
            df.set_index(index[0], inplace=True)
            groups = {
                ('value', 'all', 'all', 'all'): df.value
            }

        # Compute stats for each column, potentially grouped by year
        all_stats = []
        for g, series in groups.items():
            if g[0] != 'value':
                continue
            v, units, param, site = g
            if "year" in group or "month" in group:
                groupby = "year" if "year" in group else "month"
                dstats = self.compute_boxplots(series, groupby)
                for s in dstats:
                    s['site'] = site
                    s['parameter'] = param
                    s['units'] = units
            else:
                stats = self.compute_boxplot(series)
                stats['site'] = site
                stats['parameter'] = param
                stats['units'] = units
                dstats = [stats]
            all_stats += dstats

        df = DataFrame(all_stats)
        if "year" in group:
            index = ['year', 'site', 'parameter', 'units']
        elif "month" in group:
            index = ['month', 'site', 'parameter', 'units']
        else:
            index = ['site', 'parameter', 'units']
        df.sort(index, inplace=True)
        df.set_index(index, inplace=True)
        df.columns.name = ""
        df = df.unstack().unstack()
        if "year" in group or "month" in group:
            df = df.unstack()
        return df

    def get_grouping(self, sets):
        group = self.request.GET.get('group', None)
        if group:
            return group
        elif sets > 20:
            return "year"
        elif sets > 10:
            return "index"
        else:
            return "year-index"

    def compute_boxplots(self, series, groupby):
        def groups(d):
            if isinstance(d, tuple):
                d = d[0]
            return getattr(d, groupby)

        dstats = []
        for name, g in series.groupby(groups).groups.items():
            stats = self.compute_boxplot(series[g])
            stats[groupby] = name
            dstats.append(stats)
        return dstats

    def compute_boxplot(self, series):
        """
        Compute boxplot for given pandas Series.
        NOTE: Requires matplotlib 1.4!
        """
        from matplotlib.cbook import boxplot_stats
        series = series[series.notnull()]
        if len(series.values) == 0:
            return {}
        stats = boxplot_stats(list(series.values))[0]
        stats = {
            self.NAME_MAP.get(key, key): value
            for key, value in stats.items()
        }
        stats['count'] = len(series.values)
        stats['fliers'] = "|".join(map(str, stats['fliers']))
        return stats


class TimeSeriesView(ChartView, TimeSeriesMixin):
    pass


class ScatterView(ChartView, ScatterMixin):
    pass


class BoxPlotView(ChartView, BoxPlotMixin):
    pass



from rest_framework.serializers import Serializer
from wq.db.rest.models import get_ct, ContentType, get_object_id


class SearchResultSerializer(Serializer):
    def to_native(self, obj):
        if hasattr(obj, 'content_type_id') and hasattr(obj, 'content_object'):
            ctype = ContentType.objects.get(pk=obj.content_type_id)
            obj = obj.content_object
        else:
            ctype = get_ct(obj)

        obj_id = get_object_id(obj)

        url = ctype.urlbase
        if url:
            url += '/'
        url += obj_id

        return {
            'id': obj_id,
            'url': url,
            'type': unicode(ctype),
            'label': unicode(obj)
        }



from wq.db.rest import app
from rest_framework import serializers
from wq.db.patterns.base.serializers import TypedAttachmentSerializer
from wq.db.rest.serializers import ModelSerializer

import swapper
from wq.db.patterns.base.models import extract_nested_key
Event = swapper.load_model('vera', 'Event')
Parameter = swapper.load_model('vera', 'Parameter')
Result = swapper.load_model('vera', 'Result')


class ResultSerializer(TypedAttachmentSerializer):
    attachment_fields = ['id', 'value']
    type_model = Parameter
    value = serializers.Field()
    object_field = 'report'

    def to_native(self, obj):
        result = super(ResultSerializer, self).to_native(obj)
        if getattr(obj.type, 'units', None) is not None:
            result['units'] = obj.type.units
        has_parent = self.parent and hasattr(self.parent.opts, 'model')
        if not has_parent:
            result['report_id'] = obj.report.pk
        return result

    def from_native(self, data, files):
        obj = super(ResultSerializer, self).from_native(data, files)
        obj.value = data['value']
        return obj

    class Meta:
        exclude = ('label', 'report_id', 'report_label',
                   'value_text', 'value_numeric')


class EventSerializer(ModelSerializer):
    is_valid = serializers.Field()

    def get_default_fields(self, *args, **kwargs):
        fields = super(EventSerializer, self).get_default_fields(
            *args, **kwargs
        )
        if self.opts.depth > 0:
            Serializer = app.router.get_serializer_for_model(Result)
            fields['results'] = Serializer(context=self.context)
        return fields


class ReportSerializer(ModelSerializer):
    is_valid = serializers.Field()

    def from_native(self, data, files):
        if hasattr(data, 'dict'):
            data = data.dict()
        event_key = extract_nested_key(data, Event)
        if event_key:
            event, is_new = Event.objects.get_or_create_by_natural_key(
                *event_key
            )
            data['event'] = event.pk
        if 'request' in self.context and not data.get('user', None):
            user = self.context['request'].user
            if user.is_authenticated():
                data['user'] = user.pk
        return super(ReportSerializer, self).from_native(data, files)



from wq.db.rest import app
from django.middleware import csrf


def is_authenticated(request):
    # FIXME: Pystache contexts aren't working with Django 1.6 SimpleLazyObject
    # due to iter(data) not failing.  This makes it difficult to use {{user}}
    # in Mustache templates.  We could just override the user context variable,
    # but that will likely break Django templates, such as those used in
    # contrib.admin.  Instead, the workaround is to define "user" only within
    # the is_authenticated context.
    #
    # In Mustache templates, always use:
    #   {{#is_authenticated}}{{user.username}}{{/is_authenticated}}
    # In Django templates, just use:
    #   {{user.username}}

    if request.user.is_authenticated():
        return {
            'is_authenticated': {'user': app.router.serialize(request.user)}
        }
    else:
        return {'is_authenticated': False}


def social_auth(request):
    if (not request.user.is_authenticated()
            or not hasattr(request.user, 'social_auth')
            or request.user.social_auth.count() == 0):
        return {}

    return {
        'social_auth': {
            'accounts': [
                app.router.serialize(auth)
                for auth in request.user.social_auth.all()
            ]
        }
    }


def csrftoken(request):
    return {
        'csrftoken': csrf.get_token(request)
    }



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models
import swapper


class Migration(SchemaMigration):

    def forwards(self, orm):
        if not swapper.is_swapped('vera', 'Site'):
            # Adding model 'Site'
            db.create_table('wq_site', (
                (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
                ('latitude', self.gf('django.db.models.fields.FloatField')(null=True, blank=True)),
                ('longitude', self.gf('django.db.models.fields.FloatField')(null=True, blank=True)),
            ))
            db.send_create_signal(u'vera', ['Site'])

            # Adding unique constraint on 'Site', fields ['latitude', 'longitude']
            db.create_unique('wq_site', ['latitude', 'longitude'])

        if not swapper.is_swapped('vera', 'Event'):
            Site = swapper.load_model('vera', 'Site', orm)

            # Adding model 'Event'
            db.create_table('wq_event', (
                (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
                ('site', self.gf('django.db.models.fields.related.ForeignKey')(to=Site)),
                ('date', self.gf('django.db.models.fields.DateField')()),
            ))
            db.send_create_signal(u'vera', ['Event'])

            # Adding unique constraint on 'Event', fields ['site', 'date']
            db.create_unique('wq_event', ['site_id', 'date'])

        if not swapper.is_swapped('vera', 'Report'):
            Event = swapper.load_model('vera', 'Event', orm)
            User = swapper.load_model('auth', 'User', orm)
            ReportStatus = swapper.load_model('vera', 'ReportStatus', orm)

            # Adding model 'Report'
            db.create_table('wq_report', (
                (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
                ('event', self.gf('django.db.models.fields.related.ForeignKey')(to=Event)),
                ('entered', self.gf('django.db.models.fields.DateTimeField')(blank=True)),
                ('user', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name=u'vera_report', null=True, to=User)),
                ('status', self.gf('django.db.models.fields.related.ForeignKey')(to=ReportStatus, null=True, blank=True)),
            ))
            db.send_create_signal(u'vera', ['Report'])

        if not swapper.is_swapped('vera', 'ReportStatus'):
            # Adding model 'ReportStatus'
            db.create_table('wq_reportstatus', (
                (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
                ('slug', self.gf('django.db.models.fields.SlugField')(max_length=50)),
                ('name', self.gf('django.db.models.fields.CharField')(max_length=255)),
                ('is_valid', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ))
            db.send_create_signal(u'vera', ['ReportStatus'])

            # Adding unique constraint on 'ReportStatus', fields ['slug']
            db.create_unique('wq_reportstatus', ['slug'])

        if not swapper.is_swapped('vera', 'Parameter'):
            # Adding model 'Parameter'
            db.create_table('wq_parameter', (
                (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
                ('name', self.gf('django.db.models.fields.CharField')(max_length=255)),
                ('is_numeric', self.gf('django.db.models.fields.BooleanField')(default=False)),
                ('units', self.gf('django.db.models.fields.CharField')(max_length=50, null=True, blank=True)),
            ))
            db.send_create_signal(u'vera', ['Parameter'])

        if not swapper.is_swapped('vera', 'Result'):
            Parameter = swapper.load_model('vera', 'Parameter', orm)
            Report = swapper.load_model('vera', 'Report', orm)
            # Adding model 'Result'
            db.create_table('wq_result', (
                (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
                ('type', self.gf('django.db.models.fields.related.ForeignKey')(to=Parameter)),
                ('report', self.gf('django.db.models.fields.related.ForeignKey')(related_name='results', to=Report)),
                ('value_numeric', self.gf('django.db.models.fields.FloatField')(null=True, blank=True)),
                ('value_text', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
                ('empty', self.gf('django.db.models.fields.BooleanField')(default=False, db_index=True)),
            ))
            db.send_create_signal(u'vera', ['Result'])

            # Adding index on 'Result', fields ['type', 'report_id', 'empty']
            db.create_index('wq_result', ['type_id', 'report_id', 'empty'])

        if not swapper.is_swapped('vera', 'EventResult'):
            Event = swapper.load_model('vera', 'Event', orm)
            Result = swapper.load_model('vera', 'Result', orm)
            Site = swapper.load_model('vera', 'Site', orm)
            Parameter = swapper.load_model('vera', 'Parameter', orm)
            Report = swapper.load_model('vera', 'Report', orm)

            # Adding model 'EventResult'
            db.create_table('wq_eventresult', (
                ('id', self.gf('django.db.models.fields.PositiveIntegerField')(primary_key=True)),
                ('event', self.gf('django.db.models.fields.related.ForeignKey')(to=Event)),
                ('result', self.gf('django.db.models.fields.related.ForeignKey')(to=Result)),
                ('event_site', self.gf('django.db.models.fields.related.ForeignKey')(to=Site, null=True, blank=True)),
                ('event_date', self.gf('django.db.models.fields.DateField')()),
                ('result_type', self.gf('django.db.models.fields.related.ForeignKey')(to=Parameter)),
                ('result_report', self.gf('django.db.models.fields.related.ForeignKey')(to=Report)),
                ('result_value_numeric', self.gf('django.db.models.fields.FloatField')(null=True, blank=True)),
                ('result_value_text', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
                ('result_empty', self.gf('django.db.models.fields.BooleanField')()),
            ))
            db.send_create_signal('vera', ['EventResult'])

            # Adding unique constraint on 'EventResult', fields ['event', 'result_type']
            db.create_unique('wq_eventresult', ['event_id', 'result_type_id'])


    def backwards(self, orm):
        if not swapper.is_swapped('vera', 'EventResult'):
            # Removing unique constraint on 'EventResult', fields ['event', 'result_type']
            db.delete_unique('wq_eventresult', ['event_id', 'result_type_id'])

        if not swapper.is_swapped('vera', 'Result'):
            # Removing index on 'Result', fields ['type', 'report', 'empty']
            db.delete_index('wq_result', ['type_id', 'report_id', 'empty'])

        if not swapper.is_swapped('vera', 'ReportStatus'):
            # Removing unique constraint on 'ReportStatus', fields ['slug']
            db.delete_unique('wq_reportstatus', ['slug'])

        if not swapper.is_swapped('vera', 'Event'):
            # Removing unique constraint on 'Event', fields ['site', 'date']
            db.delete_unique('wq_event', ['site_id', 'date'])

        if not swapper.is_swapped('vera', 'Site'):
            # Removing unique constraint on 'Site', fields ['latitude', 'longitude']
            db.delete_unique('wq_site', ['latitude', 'longitude'])

        if not swapper.is_swapped('vera', 'Site'):
            # Deleting model 'Site'
            db.delete_table('wq_site')

        if not swapper.is_swapped('vera', 'Event'):
            # Deleting model 'Event'
            db.delete_table('wq_event')

        if not swapper.is_swapped('vera', 'Report'):
            # Deleting model 'Report'
            db.delete_table('wq_report')

        if not swapper.is_swapped('vera', 'ReportStatus'):
            # Deleting model 'ReportStatus'
            db.delete_table('wq_reportstatus')

        if not swapper.is_swapped('vera', 'Parameter'):
            # Deleting model 'Parameter'
            db.delete_table('wq_parameter')

        if not swapper.is_swapped('vera', 'Result'):
            # Deleting model 'Result'
            db.delete_table('wq_result')

        if not swapper.is_swapped('vera', 'EventResult'):
            # Deleting model 'EventResult'
            db.delete_table('wq_eventresult')

    models = {
        u'auth.group': {
            'Meta': {'object_name': 'Group'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        u'auth.permission': {
            'Meta': {'ordering': "(u'content_type__app_label', u'content_type__model', u'codename')", 'unique_together': "((u'content_type', u'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        u'auth.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'symmetrical': 'False', 'related_name': "u'user_set'", 'blank': 'True', 'to': u"orm['auth.Group']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'symmetrical': 'False', 'related_name': "u'user_set'", 'blank': 'True', 'to': u"orm['auth.Permission']"}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'identify.authority': {
            'Meta': {'object_name': 'Authority', 'db_table': "'wq_identifiertype'"},
            'homepage': ('django.db.models.fields.URLField', [], {'max_length': '200', 'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'object_url': ('django.db.models.fields.URLField', [], {'max_length': '200', 'null': 'True', 'blank': 'True'})
        },
        u'identify.identifier': {
            'Meta': {'object_name': 'Identifier', 'db_table': "'wq_identifier'"},
            'authority': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['identify.Authority']", 'null': 'True', 'blank': 'True'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_primary': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255', 'db_index': 'True'}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'slug': ('django.db.models.fields.SlugField', [], {'max_length': '50'})
        },
        u'vera.event': {
            'Meta': {'ordering': "('-date',)", 'unique_together': "(('site', 'date'),)", 'object_name': 'Event', 'db_table': "'wq_event'"},
            'date': ('django.db.models.fields.DateField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'site': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['vera.Site']", 'null': 'True', 'blank': 'True'})
        },
        'vera.eventresult': {
            'Meta': {'unique_together': "(('event', 'result_type'),)", 'object_name': 'EventResult', 'db_table': "'wq_eventresult'"},
            'event': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['vera.Event']"}),
            'event_date': ('django.db.models.fields.DateField', [], {}),
            'event_site': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['vera.Site']", 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.PositiveIntegerField', [], {'primary_key': 'True'}),
            'result': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['vera.Result']"}),
            'result_empty': ('django.db.models.fields.BooleanField', [], {}),
            'result_report': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['vera.Report']"}),
            'result_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['vera.Parameter']"}),
            'result_value_numeric': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'result_value_text': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'})
        },
        u'vera.parameter': {
            'Meta': {'ordering': "('name',)", 'object_name': 'Parameter', 'db_table': "'wq_parameter'"},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_numeric': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'units': ('django.db.models.fields.CharField', [], {'max_length': '50', 'null': 'True', 'blank': 'True'})
        },
        u'vera.report': {
            'Meta': {'ordering': "('-entered',)", 'object_name': 'Report', 'db_table': "'wq_report'"},
            'entered': ('django.db.models.fields.DateTimeField', [], {'blank': 'True'}),
            'event': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['vera.Event']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'status': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['vera.ReportStatus']", 'null': 'True', 'blank': 'True'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "u'vera_report'", 'null': 'True', 'to': u"orm['auth.User']"})
        },
        u'vera.reportstatus': {
            'Meta': {'unique_together': "[['slug']]", 'object_name': 'ReportStatus', 'db_table': "'wq_reportstatus'"},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_valid': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'slug': ('django.db.models.fields.SlugField', [], {'max_length': '50'})
        },
        u'vera.result': {
            'Meta': {'ordering': "('type',)", 'object_name': 'Result', 'db_table': "'wq_result'", 'index_together': "[('type', 'report', 'empty')]"},
            'empty': ('django.db.models.fields.BooleanField', [], {'default': 'False', 'db_index': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'report': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'results'", 'to': u"orm['vera.Report']"}),
            'type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['vera.Parameter']"}),
            'value_numeric': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'value_text': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'})
        },
        u'vera.site': {
            'Meta': {'unique_together': "(('latitude', 'longitude'),)", 'object_name': 'Site', 'db_table': "'wq_site'"},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'latitude': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'longitude': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'})
        },
        u'relate.relationship': {
            'Meta': {'object_name': 'Relationship', 'db_table': "'wq_relationship'"},
            'computed': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'from_content_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['contenttypes.ContentType']"}),
            'from_object_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'to_content_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['contenttypes.ContentType']"}),
            'to_object_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['relate.RelationshipType']"})
        },
        u'relate.relationshiptype': {
            'Meta': {'object_name': 'RelationshipType', 'db_table': "'wq_relationshiptype'"},
            'computed': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'from_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['contenttypes.ContentType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'inverse_name': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'to_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['contenttypes.ContentType']"})
        }
    }

    complete_apps = ['vera']



from celery import task, current_task
from xlrd import colname
from collections import namedtuple, Counter, OrderedDict
from wq.db.patterns.models import Identifier, Relationship, RelationshipType
import swapper
from .models import MetaColumn, UnknownItem, SkippedRecord, Range
from django.conf import settings
import datetime
from .signals import import_complete, new_metadata

from wq.db.rest.models import get_ct, get_object_id

import wq.db.contrib.vera.models

Site = swapper.load_model('vera', 'Site')
Event = swapper.load_model('vera', 'Event')
Report = swapper.load_model('vera', 'Report')
ReportStatus = swapper.load_model('vera', 'ReportStatus')
Parameter = swapper.load_model('vera', 'Parameter')
Result = swapper.load_model('vera', 'Result')

META_CLASSES = {
    'site': Site,
    'event': Event,
    'report': Report,
    'parameter': Parameter,
    'result': Result,
}

EVENT_KEY = [val for val, cls in Event.get_natural_key_info()]
EventKey = namedtuple('EventKey', EVENT_KEY)

CONTENT_TYPES = {
    Parameter: get_ct(Parameter),
    MetaColumn: get_ct(MetaColumn),
    UnknownItem: get_ct(UnknownItem),
}

DATE_FIELDS = {
    'DateTimeField': datetime.datetime,
    'DateField': datetime.date,
}

if hasattr(settings, 'WQ_DEFAULT_REPORT_STATUS'):
    DEFAULT_STATUS = settings.WQ_DEFAULT_REPORT_STATUS
else:
    DEFAULT_STATUS = None

PRIORITY = {
    'parameter': 1,
    'meta column': 2,
    'unknown item': 3,
}


@task
def auto_import(instance, user):
    """
    Walk through all the steps necessary to interpret and import data from an
    IO.  Meant to be called asynchronously.  Automatically suspends import if
    any additional input is needed from the user.
    """
    # Preload IO to catch any load errors early
    status = {
        'message': "Loading Data...",
        'stage': 'meta',
        'current': 1,
        'total': 4,
    }
    current_task.update_state(state='PROGRESS', meta=status)
    table = instance.load_io()

    # Parse columns
    status.update(
        message="Parsing Columns...",
        current=2,
    )
    current_task.update_state(state='PROGRESS', meta=status)
    result = read_columns(instance, user)
    if result['unknown_count']:
        result['action'] = "columns"
        result['message'] = "Input Needed"
        return result

    # Parse row identifiers
    status.update(
        message="Parsing Identifiers...",
        current=3,
    )
    current_task.update_state(state='PROGRESS', meta=status)
    result = read_row_identifiers(instance, user)
    if result['unknown_count']:
        result['action'] = "ids"
        result['message'] = "Input Needed"
        return result

    status.update(
        message="Importing Data...",
        current=4,
    )

    # The rest is the same as import_data
    return do_import(instance, user)


def get_choices(instance):
    def make_list(cls, name):
        rows = cls.objects.all()
        ct = CONTENT_TYPES[cls]
        result = [{
            'url': '%s/%s' % (ct.urlbase, get_object_id(row)),
            'label': unicode(row)
        } for row in rows]
        result.insert(0, {
            'url': '%s/new' % ct.urlbase,
            'label': "New %s" % name,
        })
        return {
            'name': name,
            'choices': result
        }

    return [
        make_list(MetaColumn, "Metadata Column"),
        make_list(Parameter, "Parameter")
    ]


@task
def read_columns(instance, user=None):
    matched = get_columns(instance)
    unknown_count = 0
    for info in matched:
        if info['type'] == 'unknown':
            unknown_count += 1
            # Add some useful context items for client
            info['unknown'] = True
            info['types'] = get_choices(instance)

    return {
        'columns': matched,
        'unknown_count': unknown_count,
    }


# FIXME: These functions might make more sense as methods on IoModel
def get_columns(instance):
    if instance.already_parsed():
        return load_columns(instance)
    else:
        return parse_columns(instance)


def get_meta_columns(instance):
    matched = get_columns(instance)
    cols = OrderedDict()
    for col in matched:
        if 'colnum' not in col or not col['type']:
            continue
        if col['type'] in ('parameter_value', 'unknown'):
            continue
        if col['type'] in ('event', 'report', 'result'):
            continue
        cols.setdefault(col['type'], [])
        cols[col['type']].append(col)
    return cols


def load_columns(instance):
    rels = instance.relationships.filter(type__name='Contains Column')
    table = instance.load_io()

    matched = []
    for rel in rels:
        item = rel.right
        info = {
            'match': unicode(item),
            'rel_id': rel.pk,
        }
        if isinstance(item, UnknownItem):
            info['type'] = "unknown"
            info['value'] = item.name
        elif isinstance(item, Parameter):
            info['type'] = "parameter_value"
            info['parameter_id'] = get_object_id(item)
        elif isinstance(item, MetaColumn):
            info['type'] = item.type
            info['field_name'] = item.name

        if rel.range_set.filter(type='list').exists():
            col = rel.range_set.get(type='list').start_column
            info['name'] = table.field_map.keys()[col].replace('\n', ' - ')
            info['column'] = colname(col)
            info['colnum'] = col

        elif rel.range_set.filter(type='value').exists():
            info['name'] = get_range_value(table, rel.range_set.get(
                type='head'
            ))
            info['value'] = get_range_value(table, rel.range_set.get(
                type='value'
            ))
        matched.append(info)
    matched.sort(key=lambda info: info.get('colnum', -1))
    return matched


def get_range_value(table, rng):
    if rng.start_row == rng.end_row and rng.start_column == rng.end_column:
        return table.extra_data.get(rng.start_row, {}).get(rng.start_column)

    val = u""
    for r in range(rng.start_row, rng.end_row + 1):
        for c in range(rng.start_column, rng.end_column + 1):
            val += unicode(table.extra_data.get(r, {}).get(c, ""))
    return val


def parse_columns(instance):
    table = instance.load_io()
    if table.tabular:
        for r in table.extra_data:
            row = table.extra_data[r]
            for c in row:
                if c + 1 in row and c - 1 not in row:
                    parse_column(
                        instance,
                        name=row[c],
                        head=[r, c, r, c],
                        body=[r, c + 1, r, c + 1],
                        body_type='value'
                    )

    for i, name in enumerate(table.field_map.keys()):
        if table.tabular:
            header_row = table.header_row
            start_row = table.start_row
        else:
            header_row = -1
            start_row = 0
        name = table.clean_field_name(name)
        parse_column(
            instance,
            name=name,
            head=[header_row, i, start_row - 1, i],
            body=[start_row, i, start_row + len(table), i],
            body_type='list'
        )

    return load_columns(instance)


def parse_column(instance, name, head, body, body_type):
    matches = list(Identifier.objects.filter_by_identifier(name))
    if len(matches) > 0:
        matches.sort(
            key=lambda ident: PRIORITY.get(ident.content_type.name, 0)
        )
        column = matches[0].content_object
        ctype = matches[0].content_type
    else:
        column = UnknownItem.objects.find(name)
        ctype = CONTENT_TYPES[UnknownItem]

    reltype, is_new = RelationshipType.objects.get_or_create(
        from_type=get_ct(instance),
        to_type=ctype,
        name='Contains Column',
        inverse_name='Column In'
    )
    rel = instance.relationships.create(
        type=reltype,
        to_content_type=ctype,
        to_object_id=column.pk,
    )
    Range.objects.create(
        relationship=rel,
        type='head',
        start_row=head[0],
        start_column=head[1],
        end_row=head[2],
        end_column=head[3]
    )
    Range.objects.create(
        relationship=rel,
        type=body_type,
        start_row=body[0],
        start_column=body[1],
        end_row=body[2],
        end_column=body[3]
    )


@task
def update_columns(instance, user, post):
    matched = get_columns(instance)
    for col in matched:
        if col['type'] != 'unknown':
            continue
        val = post.get('rel_%s' % col['rel_id'], None)
        if not val:
            continue

        vtype, vid = val.split('/')
        if vtype == 'parameters':
            cls = Parameter
        elif vtype == 'metacolumns':
            cls = MetaColumn
        else:
            continue

        if vid == 'new':
            obj = cls.objects.find(col['value'])
            obj.contenttype = CONTENT_TYPES[Parameter]
            obj.save()
            ident = obj.primary_identifier
        else:
            obj = cls.objects.get_by_identifier(vid)
            ident = obj.identifiers.create(
                name=col['value']
            )

        reltype, is_new = RelationshipType.objects.get_or_create(
            from_type=get_ct(instance),
            to_type=CONTENT_TYPES[cls],
            name='Contains Column',
            inverse_name='Column In'
        )
        rel = instance.relationships.get(pk=col['rel_id'])
        rel.type = reltype
        rel.right = obj
        rel.save()

        new_metadata.send(
            sender=update_columns,
            instance=instance,
            object=obj,
            identifier=ident,
        )

    return read_columns(instance)


@task
def read_row_identifiers(instance, user):
    coltypes = get_meta_columns(instance)
    ids = {}
    for mtype in coltypes:
        ids[mtype] = Counter()

    for row in instance.load_io():
        for mtype, cols in coltypes.items():
            counter = ids[mtype]
            meta = OrderedDict()
            for col in cols:
                meta[col['field_name']] = row[col['colnum']]
            key = tuple(meta.items())
            counter[key] += 1

    idgroups = []
    unknown_ids = 0
    for mtype in ids:
        cls = META_CLASSES[mtype]
        idinfo = {
            'type_id': mtype,
            'type_label': mtype.capitalize(),
            'ids': []
        }
        for key, count in ids[mtype].items():
            meta = OrderedDict(key)
            try:
                obj = cls.objects.get_by_natural_key(meta['id'])
            except cls.DoesNotExist:
                obj = None

            info = {
                'value': meta['id'],
                'count': count,
                'meta': [{
                    'name': k,
                    'value': v
                } for k, v in meta.items() if k != 'id']
            }
            if obj is not None:
                info['match'] = unicode(obj)
                # FIXME: Confirm that metadata hasn't changed
            else:
                info['ident_id'] = unknown_ids
                unknown_ids += 1
                info['unknown'] = True
                choices = instance.get_id_choices(cls, meta)
                info['choices'] = [{
                    'id': get_object_id(obj),
                    'label': unicode(obj),
                } for obj in choices]
                info['choices'].insert(0, {
                    'id': 'new',
                    'label': "New %s" % idinfo['type_label'],
                })

            idinfo['ids'].append(info)
        idinfo['ids'].sort(key=lambda info: info['value'])
        idgroups.append(idinfo)

    if not unknown_ids:
        # Create relationships after all IDs are resolved.

        # FIXME: parse_columns() always creates relationships, using
        # UnknownItem to stand in for unknown column identifiers.
        # Use UnknownItem here as well?

        from_type = get_ct(instance)

        for idinfo in idgroups:
            cls = META_CLASSES[idinfo['type_id']]
            to_type = get_ct(cls)
            reltype, is_new = RelationshipType.objects.get_or_create(
                from_type=from_type,
                to_type=to_type,
                name='Contains Identifier',
                inverse_name='Identifier In'
            )
            for info in idinfo['ids']:
                obj = cls.objects.get_by_natural_key(info['value'])
                rel, isnew = instance.relationships.get_or_create(
                    type=reltype,
                    to_content_type=to_type,
                    to_object_id=obj.pk,
                    from_content_type=from_type,
                    from_object_id=instance.pk,
                )

                # FIXME: create equivalent of Range objects here?

    return {
        'unknown_count': unknown_ids,
        'types': idgroups,
    }


@task
def update_row_identifiers(instance, user, post):
    coltypes = get_meta_columns(instance)
    unknown_count = int(post.get('unknown_count', 0))
    for i in range(0, unknown_count):
        ident_value = post.get('ident_%s_value' % i, None)
        ident_type = post.get('ident_%s_type' % i, None)
        ident_id = post.get('ident_%s_id' % i, None)
        if ident_value and ident_type and ident_id:
            if ident_type not in coltypes:
                raise Exception("Unexpected type %s!" % ident_type)
            cls = META_CLASSES[ident_type]
            if ident_id == 'new':
                # Create new object (with primary identifier)
                obj = cls.objects.find(ident_value)
                ident = obj.primary_identifier

                # Set additional metadata fields on new object
                for col in coltypes[ident_type]:
                    if col['field_name'] == "id":
                        continue

                    formname = 'ident_%s_%s' % (i, col['field_name'])
                    val = post.get(formname, None)
                    if not val:
                        continue

                    setattr(obj, col['field_name'], val)

                    # If present, also apply "name" attribute to identifier
                    if col['field_name'] == "name":
                        ident.name = val
                        ident.slug = ident_value
                        ident.save()
                obj.save()
            else:
                # Add new identifier to existing object for future reference
                obj = cls.objects.get_by_identifier(ident_id)
                name = post.get('ident_%s_name' % i, None)
                if not name:
                    name = ident_value
                ident = obj.identifiers.create(name=name, slug=ident_value)

                # FIXME: Update existing metadata?

            new_metadata.send(
                sender=update_row_identifiers,
                instance=instance,
                object=obj,
                identifier=ident,
            )

    return read_row_identifiers(instance, user)


@task
def reset(instance, user):
    instance.relationships.all().delete()


@task
def import_data(instance, user):
    """
    Import all parseable data from the dataset instance's IO class.
    """
    return do_import(instance, user)


def do_import(instance, user):

    # (Re-)Load data and column information
    table = instance.load_io()
    matched = get_columns(instance)

    # Set global defaults for metadata values
    if not user.is_authenticated():
        user = None
    instance_globals = {
        # Metadata fields
        'event_meta': {},
        'report_meta': {
            'user': user,
            'status_id': DEFAULT_STATUS,
        },
        # result_meta, parameter_meta, and site_meta are defined on a
        # case-by-case basis; see create_record() below

        # Result values indexed by parameter (for "horizontal" tables)
        'param_vals': {}
    }

    # Set default to None for any event key fields that are not required
    for field_name in EVENT_KEY:
        info = Event._meta.get_field_by_name(field_name)
        field = info[0]
        if field.null:
            instance_globals['event_meta'][field_name] = None

    # Set any global defaults defined within data themselves (usually as extra
    # cells above the headers in a spreadsheet)
    for col in matched:
        if 'value' in col:
            save_value(col, col['value'], instance_globals)

    # Loop through table rows and add each record
    rows = len(table)
    errors = []
    skipped = []

    if table.tabular:
        def rownum(i):
            return i + table.start_row
    else:
        def rownum(i):
            return i

    for i, row in enumerate(table):
        # Update state (for status() on view)
        current_task.update_state(state='PROGRESS', meta={
            'message': "Importing Data...",
            'stage': 'data',
            'current': i + 1,
            'total': rows,
            'skipped': skipped
        })

        # Create report, capturing any errors
        skipreason = None
        try:
            report = create_report(row, instance_globals, matched)
        except Exception as e:
            # Log error in database
            skipreason = repr(e)
            skipped.append({'row': rownum(i) + 1, 'reason': skipreason})
            report, is_new = SkippedRecord.objects.get_or_create(
                reason=skipreason
            )

        # Record relationship between data source and resulting report (or
        # skipped record), including specific cell range.
        rel = instance.create_relationship(
            report,
            'Contains Row',
            'Row In'
        )
        Range.objects.create(
            relationship=rel,
            type='row',
            start_row=rownum(i),
            start_column=0,
            end_row=rownum(i),
            end_column=len(row) - 1
        )

    # Send completion signal (in case any server handlers are registered)
    status = {
        'current': i + 1,
        'total': rows,
        'skipped': skipped
    }
    import_complete.send(sender=import_data, instance=instance, status=status)

    # Return status (thereby updating task state for status() on view)
    return status


def create_report(row, instance_globals, matched):
    """
    Create actual report instance from parsed values.
    """

    # Copy global values to record hash
    record = {
        key: instance_globals[key].copy()
        for key in instance_globals
    }

    # In some spreadsheets (i.e. "horizontal" tables), multiple columns
    # indicate parameter names and each row only contains result values.  In
    # others (i.e. "vertical" tables), each row lists both the parameter name
    # and the value.  See http://wq.io/docs/dbio#horizontal-vs-vertical

    # Parse metadata & "horizontal" table parameter values
    for col in matched:
        if 'colnum' in col:
            save_value(col, row[col['colnum']], record)

    # Handle "vertical" table values (parsed as metadata by save_value())
    if 'result_meta' in record and 'parameter_meta' in record:
        # FIXME: handle other parameter & result metadata
        parameter_id = record['parameter_meta']['id']
        result_value = record['result_meta']['value']
        record['param_vals'][parameter_id] = result_value

    if 'site_meta' in record:
        # FIXME: Handle other site metadata
        site_id = record['site_meta']['id']
        record['event_meta']['site'] = site_id

    # Ensure complete Event natural key (http://wq.io/docs/erav#natural-key)
    missing = set(EVENT_KEY) - set(record['event_meta'].keys())
    if missing:
        raise Exception(
            'Incomplete Record - missing %s' % ", ".join(missing)
        )

    # Create report instance
    report = Report.objects.create_report(
        EventKey(**record['event_meta']),
        record['param_vals'],
        **record['report_meta']
    )
    return report


def save_value(col, val, obj):
    """
    For each cell in each row, use parsed col(umn) information to determine how
    to apply the cell val(ue) to the obj(ect hash).
    """
    if col['type'] == "parameter_value":
        # Parameter value in a "horizontal" table
        save_parameter_value(col, val, obj)
    elif col['type'] != "unknown":
        # Metadata value in either a "horizontal" or "vertical" table
        save_metadata_value(col, val, obj)


def save_parameter_value(col, val, obj):
    """
    This column was identified as a parameter; update param_vals with the
    cell value from this row.
    """
    parameter_id = col['parameter_id']
    vals = obj['param_vals']

    # Hack: if there are two (or more) columns pointing to the same
    # parameter, join both values together with a space.
    if parameter_id in vals:
        val = "%s %s" % (vals[parameter_id], val)

    vals[parameter_id] = val


def save_metadata_value(col, val, obj):
    """
    This column was identified as a metadata field; update the metadata
    for the related object with the cell value from this row.
    """

    # Skip empty values
    if val is None or val == '' or not col['type']:
        return

    # Assign metadata property based on meta_field (MetaColumn.name).
    meta_key = '%s_meta' % col['type']
    meta_cls = META_CLASSES[col['type']]
    meta_field = col['field_name']

    # Event and report metadata are defined whether this is a "horizontal" or
    # "vertical" table.  On the other hand, parameter/result meta are unique to
    # "vertical" tables and are thus not included by default.  In either case,
    # site metadata is also not expected.
    if meta_key not in obj:
        # This is one of parameter, result, or site meta.
        obj[meta_key] = {}

    # A meta_field value of '[field].[part]' indicates a value is split across
    # multiple columns.  For example, a spreadsheet could contain two columns
    # (date and time) that would be merged into a single "observed" field on a
    # custom Event class.  There would then be two MetaColumns values, with
    # names of "observed.date" and "observed.time" respectively.
    if '.' in meta_field:
        meta_field, part = meta_field.split('.')
    else:
        part = None

    # Determine Django field type (for metadata models only)
    if col['type'] == 'result':
        meta_datatype = None
    else:
        meta_datatype = meta_cls._meta.get_field_by_name(
            meta_field,
        )[0].get_internal_type()

    # Automatically parse date values as such
    if (meta_datatype in DATE_FIELDS and isinstance(val, basestring)
            and part != 'time'):
        from dateutil.parser import parse
        val = parse(val)
        if meta_datatype == 'DateField':
            val = val.date()

    # If field is already set by an earlier column, this value might be the
    # second half of a date/time pair.
    if obj[meta_key].get(meta_field, None) is not None:
        if not part:
            raise Exception(
                'Multiple columns found for %s' % meta_field
            )
        if part not in ('date', 'time'):
            raise Exception(
                'Unexpected multi-column field name: %s.%s!' % (
                    meta_field, part
                )
            )
        other_val = obj[meta_key][meta_field]
        val = process_date_part(val, other_val, part)

    # Save value to parsed record
    obj[meta_key][meta_field] = val


def process_date_part(new_val, old_val, part):
    """
    Combine separate date & time columns into a single value.
    """

    if part == 'date':
        date, time = new_val, old_val
    else:
        date, time = old_val, new_val

    # Date should already be a valid date (see parse in save_metadata_value)
    if not isinstance(date, datetime.date):
        raise Exception("Expected date but got %s!" % date)

    # Try some extra hacks to convert time values
    if not isinstance(time, datetime.time):
        if (isinstance(time, float)
                and time >= 100 and time <= 2400):
            # "Numeric" time (hour * 100 + minutes)
            time = str(time)
        elif isinstance(time, basestring) and ":" in time:
            # Take out semicolon for isdigit() code below
            time = time.replace(":", "")

        # FIXME: what about seconds?
        if time.isdigit() and len(time) in (3, 4):
            if len(time) == 3:
                # 300 -> time(3, 0)
                time = datetime.time(
                    int(time[0]),
                    int(time[1:])
                )
            else:
                # 1200 -> time(12, 0)
                time = datetime.time(
                    int(time[0:2]),
                    int(time[2:])
                )
        else:
            # Meh, it was worth a shot
            raise Exception("Expected time but got %s!" % time)
    return datetime.datetime.combine(date, time)



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import DataMigration
from django.db import models
import swapper


class Migration(DataMigration):

    def forwards(self, orm):
        if swapper.is_swapped('annotate', 'AnnotationType'):
            return

        if swapper.is_swapped('annotate', 'Annotation'):
            return

        if db.backend_name != 'postgres':
            print "Warning: Non-postgres database detected; convenience view will not be created."
            return

        db.execute('''
CREATE OR REPLACE VIEW wq_annotation_joined AS
SELECT ct.app_label, ct.model, a.object_id, at.id AS type_id, at.name AS type_name, a.value
FROM wq_annotation a
JOIN wq_annotationtype at ON at.id = a.type_id
JOIN django_content_type ct ON a.content_type_id = ct.id;''')

    def backwards(self, orm):
        if swapper.is_swapped('annotate', 'AnnotationType'):
            return
        if swapper.is_swapped('annotate', 'Annotation'):
            return
        if db.backend_name != 'postgres':
            return
        db.execute("DROP VIEW wq_annotation_joined;")

    models = {
        'annotate.annotation': {
            'Meta': {'object_name': 'Annotation', 'db_table': "'wq_annotation'"},
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['annotate.AnnotationType']"}),
            'value': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'})
        },
        'annotate.annotationtype': {
            'Meta': {'object_name': 'AnnotationType', 'db_table': "'wq_annotationtype'"},
            'contenttype': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']", 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        }
    }

    complete_apps = ['annotate']



from django.contrib.gis.db import models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic
from wq.db.patterns.base import SerializableGenericRelation

from django.conf import settings
SRID = getattr(settings, 'SRID', 4326)


class Location(models.Model):
    name = models.CharField(max_length=255, null=True, blank=True)
    is_primary = models.BooleanField(default=False)
    geometry = models.GeometryField(srid=SRID)
    accuracy = models.IntegerField(null=True, blank=True)

    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    content_object = generic.GenericForeignKey()

    objects = models.GeoManager()

    def __unicode__(self):
        if self.name is not None and len(self.name) > 0:
            return '%s - %s' % (self.name, self.content_object)
        else:
            return 'Location %s - %s' % (self.pk, self.content_object)

    class Meta:
        db_table = 'wq_location'


class LocatedModel(models.Model):
    locations = SerializableGenericRelation(Location)

    class Meta:
        abstract = True



from wq.db.rest import app
from .views import LoginView, LogoutView, UserViewSet
from django.contrib.auth.models import User

app.router.add_page('login', {'url': 'login'}, LoginView)
app.router.add_page('logout', {'url': 'logout'}, LogoutView)

app.router.register_viewset(User, UserViewSet)



# Use Django's template loader with Pystache's renderer

from pystache.renderer import Renderer as PystacheRenderer
from django.template.loaders.filesystem import Loader as FileSystemLoader


class Renderer(PystacheRenderer):
    def get_template_source(self, name):
        template, origin = Loader().load_template(name)
        return template.source

    def _make_load_template(self):
        return self.get_template_source

    def _make_load_partial(self):
        def load_partial(name):
            return self.get_template_source('partials/' + name + '.html')
        return load_partial

    def str_coerce(self, val):
        if val is None:
            return ""
        return str(val)

renderer = Renderer()


class Template(object):
    def __init__(self, source):
        #save raw source for actual rendering by Pystache
        self.source = source

    def render(self, context):
        return renderer.render(self.source, *context)


class Loader(FileSystemLoader):
    def load_template(self, template_name, template_dirs=None):
        source, origin = self.load_template_source(
            template_name, template_dirs
        )
        template = Template(source)
        return template, origin



from django.conf.urls import patterns, include, url
from django.conf import settings

from .views import SearchView, DisambiguateView

urlpatterns = patterns('',
    url('^search/?$', SearchView.as_view()),
    url('^search\.(?P<format>\w+)$', SearchView.as_view())
)

if getattr(settings, 'DISAMBIGUATE', False):
    urlpatterns += patterns('',
        url('^(?P<slug>[^\/]+)$', DisambiguateView.as_view())
    )
